from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest
from . import adskalman
import numpy
import pkg_resources
import warnings

def assert_3d_vs_kpm_close(A,B,debug=False):
    """compare my matrices (where T dimension is first) vs. KPM's (where it's last)"""
    for i in range(A.shape[0]):
        if debug:
            print()
            print('i=%d'%i)
            print('A[i],',A[i].shape,A[i])
            print("B[:,:,i]",B[:,:,i].shape,B[:,:,i])
            diff = A[i]-B[:,:,i]
            print('diff',diff)
        try:
            assert numpy.allclose( A[i], B[:,:,i])
            if debug:
                print('(same)')
        except Exception as err:
            if debug:
                print('(different)')
            raise

class TestStats(unittest.TestCase):
    def test_likelihood1(self):
        pi = numpy.pi
        exp = numpy.exp

        x=numpy.array([0,0])
        m=numpy.array([[0,0]])
        C=numpy.array([[1.0,0],[0,1]])
        lik = adskalman.gaussian_prob(x,m,C)
        lik_should = 1/(2*pi) * 1 * exp( 0 )
        assert numpy.allclose(lik,lik_should)

    def test_rand_mvn1(self):
        mu = numpy.array([1,2,3,4])
        sigma = numpy.eye(4)
        sigma[:2,:2] = [[2.0, 0.1],[0.1,0.2]]
        N = 1000
        Y = adskalman.rand_mvn(mu,sigma,N)
        assert Y.shape==(N,4)
        mu2 = numpy.mean(Y,axis=0)

        eps = .2
        assert numpy.sqrt(numpy.sum((mu-mu2)**2)) < eps # expect occasional failure here

        sigma2 = adskalman.covar(Y)
        eps = .3
        assert numpy.sqrt(numpy.sum((sigma-sigma2)**2)) < eps # expect occasional failure here

    def test_rand_mvn2(self):
        mu = numpy.array([1])
        sigma = 2*numpy.eye(1)
        N = 10000
        Y = adskalman.rand_mvn(mu,sigma,N)
        Y2 = numpy.random.standard_normal((N,1))*sigma[0]+mu[0]
        assert Y.shape == (N,1)
        assert Y2.shape == (N,1)

        Y = Y[:,0] # drop an axis
        Y2 = Y2[:,0] # drop an axis

        meanY = numpy.mean(Y)
        #print 'meanY',meanY
        meanY2 = numpy.mean(Y2)
        #print 'meanY2',meanY2

        # according to z-test on wikipedia
        SE = sigma[0]/numpy.sqrt(N)
        SE2 = sigma[0]/numpy.sqrt(N)

        z = (meanY-mu[0])/SE
        z2 = (meanY2-mu[0])/SE2

        if abs(z) > 3:
            raise ValueError('Sample mean was three sigma away from true mean. This error is expected rarely.')
        if abs(z2) > 3:
            raise ValueError('Sample mean was three sigma away from true mean. This error is expected rarely.')
        #print 'z',z
        #print 'z2',z2

##         assert abs(meanY) < 1e-5

##         est_std_diff = abs(numpy.std(Y)-1)
##         #print 'err',est_std_diff
##         assert est_std_diff < tol

class TestAutogeneratedData(unittest.TestCase):
    # parametric.ParametricTestCase
    #: Prefix for tests with independent state.  These methods will be run with
    #: a separate setUp/tearDown call for each test in the group.
    #_indepParTestPrefix = 'test_kalman_murphy_smooth'

    def setUp(self):
        # process model
        A = numpy.array([[1, 0],
                         [0, .5]],
                        dtype=numpy.float64)

        # observation model
        C = numpy.array([[1, 0],
                         [0, 1]],
                        dtype=numpy.float64)

        # process covariance
        Q = numpy.eye(2)

        # measurement covariance
        R = numpy.eye(2)

        Q = 0.001*Q
        R = 0.001*R

        N = 1000
        numpy.random.seed(4)
        x0 = numpy.random.randn( 2 )
        P0 = numpy.random.randn( 2,2 )

        process_noise = adskalman.rand_mvn( 0, Q, N )
        observation_noise = adskalman.rand_mvn( 0, R, N )

        X = []
        Y = []
        x = x0
        for i in range(N):
            X.append( x )
            Y.append( numpy.dot(C,x) + observation_noise[i] )

            x = numpy.dot(A,x) + process_noise[i] # for next time step

        X = numpy.array(X)
        Y = numpy.array(Y)

        Abad = numpy.eye(2)
        Abad[0,0] = .1
        Abad[1,1] = 4

        self.X = X
        self.Y = Y
        self.A = A
        self.Abad = Abad
        self.C = C
        self.Q = Q
        self.R = R
        self.x0 = x0
        self.P0 = P0

    def test_kalman_murphy_smooth(self):
        test = self._test_kalman_murphy_smooth
        for missing_type in ['nan','large']:
            #yield (test,missing_type)
            test(missing_type)

    def _test_kalman_murphy_smooth(self,missing_type):
        Ymissing = numpy.array(self.Y, copy=True)

        if missing_type=='nan':
            valid_idx = None
        elif missing_type=='large':
            valid_idx = list(range(len(Ymissing)))
        else:
            raise ValueError('unknown missing_type "%s"'%missing_type)

        bad_idx = list(range(5,len(Ymissing),10))
        for i in bad_idx:
            if missing_type=='nan':
                Ymissing[i] = numpy.nan*Ymissing[i]
            elif missing_type=='large':
                valid_idx.remove(i)
                Ymissing[i] = 9e999*Ymissing[i]

        assert len(bad_idx)>5 # make sure some data are actually missing!
        xsmooth, Vsmooth = adskalman.kalman_smoother(Ymissing,self.A,self.C,self.Q,self.R,self.x0,self.P0,
                                                     valid_data_idx=valid_idx)
        dist = numpy.sqrt(numpy.sum((xsmooth-self.X)**2,axis=1))
        mean_dist = numpy.mean(dist)
        #print 'mean_dist',missing_type,mean_dist
        assert mean_dist < 0.1 # Not sure of exact value, but shouldn't be too large

    def _test_kalman_murphy_EM(self): # temporarily disabled
        def returnC(orig):
            return self.C
        def returnQ(orig):
            return self.Q
        def returnR(orig):
            return self.R
        constr_fun_dict = {
            'C':returnC,
            #'Q':returnQ,
            #'R':returnR,
            }
        if 1:
            print('-'*80,'Murphy')
        Alearn, Clearn, Qlearn, Rlearn, initx2, initV2, LL = adskalman.learn_kalman(
            self.Y, self.Abad, self.C, self.Q, self.R, self.x0,
            self.P0, max_iter=500, verbose=2,
            constr_fun_dict=constr_fun_dict,
            thresh=1e-5,
            )

        xsmooth, Vsmooth, VVsmooth, loglik = adskalman.kalman_smoother(
            self.Y,self.A,self.C,self.Q,self.R,self.x0,self.P0,
            full_output=True)
        if 1:
            print('best LL',loglik)

            print()
            print('*'*80,'Murphy')
            print('A')
            print(self.A)
            print('Alearn')
            print(Alearn)
            print('Clearn')
            print(Clearn)
            print('Qlearn')
            print(Qlearn)
            print('Rlearn')
            print(Rlearn)
            print('*'*80,'Murphy')

class TestLotsOfAutogeneratedData(unittest.TestCase):
    def setUp(self):
        dt = 0.5
        # process model
        A = numpy.array([[1, 0, dt, 0],
                         [0, 1, 0, dt],
                         [0, 0, .5,  0],
                         [0, 0, 0,  .5]],
                        dtype=numpy.float64)
        # observation model
        C = numpy.array([[1, 0, 0, 0],
                         [0, 1, 0, 0]],
                        dtype=numpy.float64)

        # process covariance
        Q = numpy.eye(4)
        Q[:2,:2] = [[ 2, .1],[.1,2]]

        # measurement covariance
        R = numpy.array([[ 1, 0.2], [0.2, 1]],
                        dtype=numpy.float64)

        Q = 0.001*Q
        R = 0.001*R

        N = 1000
        numpy.random.seed(4)
        x0 = numpy.random.randn( 4 )
        P0 = numpy.random.randn( 4,4 )

        process_noise = adskalman.rand_mvn( 0, Q, N )
        observation_noise = adskalman.rand_mvn( 0, R, N )

        X = []
        Y = []
        x = x0
        for i in range(N):
            X.append( x )
            Y.append( numpy.dot(C,x) + observation_noise[i] )

            x = numpy.dot(A,x) + process_noise[i] # for next time step

        X = numpy.array(X)
        Y = numpy.array(Y)

        Abad = numpy.eye(4)
        #Abad[0,0] = 0
        #Abad[1,1] = 100
        #Abad[0,2] = .0
        #Abad[1,3] = .0

        self.Y = Y
        self.A = A
        self.Abad = Abad
        self.C = C
        self.Q = Q
        self.R = R
        self.x0 = x0
        self.P0 = P0

    def _test_kalman_ads1(self):
        print('-'*80,'DRO')
        xlearn, Vlearn, Alearn, Clearn, Qlearn, Rlearn = adskalman.DROsmooth(self.Y,
                                                                             self.Abad,
                                                                             self.C,
                                                                             self.Q,
                                                                             self.R,
                                                                             self.x0,
                                                                             self.P0,
                                                                             mode='EM')
        print()
        print('*'*80,'DRO')
        print('Alearn',Alearn)
        print('Clearn',Clearn)
        print('Qlearn',Qlearn)
        print('Rlearn',Rlearn)
        print('*'*80,'DRO')


class TestKalman(unittest.TestCase):
    def test_kalman1(self,time_steps=100,Qsigma=0.1,Rsigma=0.5):
        dt = 0.1
        # process model
        A = numpy.array([[1, 0, dt, 0],
                         [0, 1, 0, dt],
                         [0, 0, 1,  0],
                         [0, 0, 0,  1]],
                        dtype=numpy.float64)
        # observation model
        C = numpy.array([[1, 0, 0, 0],
                         [0, 1, 0, 0]],
                        dtype=numpy.float64)
        # process covariance
        Q = Qsigma*numpy.eye(4)
        # measurement covariance
        R = Rsigma*numpy.eye(2)

        x = numpy.array([0,0,0,0],dtype=numpy.float)
        x += Qsigma*numpy.random.standard_normal(size=x.shape)

        kf = adskalman.KalmanFilter(A,C,Q,R,x,Q)
        y = numpy.dot(C,x)
        y += Rsigma*numpy.random.standard_normal(y.shape)

        xs = []
        xhats = []
        for i in range(time_steps):
            if i==0: isinitial=True
            else: isinitial = False

            xhat,P = kf.step(y=y, isinitial=isinitial)

            # calculate new state
            x = numpy.dot(A,x) + Qsigma*numpy.random.standard_normal(x.shape)
            # and new observation
            y = numpy.dot(C,x) + Rsigma*numpy.random.standard_normal(y.shape)

            xs.append(x)
            xhats.append( xhat )
        xs = numpy.array(xs)
        xhats = numpy.array(xhats)
        # XXX some comparison between xs and xhats

def get_test_suite():
    ts=unittest.TestSuite([unittest.makeSuite(TestKalman),
                           unittest.makeSuite(TestAutogeneratedData),
                           #unittest.makeSuite(TestLotsOfAutogeneratedData),
                           unittest.makeSuite(TestStats),
                           ])
    return ts

if __name__=='__main__':
    numpy.set_printoptions(suppress=True)
    if 0:
        suite = get_test_suite()
        suite.debug()
    elif 0:
        suite = unittest.makeSuite(TestLotsOfAutogeneratedData)
        suite.debug()
    elif 1:
        suite = unittest.makeSuite(TestAutogeneratedData)
        suite.debug()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy
import numpy.matlib
linalg = numpy.linalg
# import scikits.learn.machine.em.densities as densities

# For treatment of missing data, see:
#
# Shumway, R.H. & Stoffer, D.S. (1982). An approach to time series
# smoothing and forecasting using the EM algorithm. Journal of Time
# Series Analysis, 3, 253-264. http://www.stat.pitt.edu/stoffer/em.pdf


def rand_mvn(mu, sigma, N):
    """generate multivariate normal samples
input:
  mu - a length M vector, the mean of the distribution
  sigma - an M by M array, the covariance of the distribution
  N - a scalar N, the number of samples to generate
output:
  Y - an N by M array, N samples with mean mu and covariance sigma
"""
    M = sigma.shape[0]
    X = numpy.random.standard_normal((M, N))
    Y = numpy.dot(linalg.cholesky(sigma), X).T + mu
    return Y


def covar(x):
    """determine the sample covariance matrix of x

input:
  x - an N by M array, N samples of an M component vector
output:
  sigma - and M by M array, the covariance matrix
"""
    mu = numpy.mean(x, axis=0)
    N = x.shape[0]
    y = x - mu
    sigma = numpy.dot(y.T, y) / (N - 1)

    # Note, the maximum likelihood estimator is /N [not /(N-1)] as
    # above, but that works only for a multivariate normal.

    return sigma


def gaussian_prob(x, m, C, use_log=False):
    if 0:
        return numpy.asscalar(densities.gauss_den(x, m, C, log=use_log))
    # Kevin Murphy's implementation
    m = numpy.atleast_1d(m)
    assert x.ndim == 1
    N = 1
    d = x.shape[0]
    M = numpy.dot(m.T, numpy.ones((1, N)))  # replicate mean across columns
    denom = (2 * numpy.pi)**(d / 2) * numpy.sqrt(abs(numpy.linalg.det(C)))
    x = x[:, numpy.newaxis]  # make column vector
    xMT = (x - M).T
    tmpXX = (numpy.dot(xMT, numpy.linalg.inv(C))) * xMT
    mahal = numpy.sum(tmpXX.flat)
    if numpy.any(mahal < 0):
        raise ValueError("mahal < 0 => C is not psd")
    if use_log:
        p = -0.5 * mahal - numpy.log(denom)
    else:
        eps = numpy.finfo(numpy.float64).eps
        # eps=2**-52
        p = numpy.exp(-0.5 * mahal) / (denom + eps)
    return p


class VariableObservationNoiseKalmanFilter:

    def __init__(self, A, C, Q, initial_x, initial_P):
        ss = len(A)  # ndim in state space
        os = len(C)  # ndim in observation space

        assert A.shape == (ss, ss)
        assert C.shape == (os, ss)
        assert Q.shape == (ss, ss)
        assert initial_x.shape == (ss,)
        assert initial_P.shape == (ss, ss)

        self.A = A  # process update model
        self.C = C  # observation model
        self.Q = Q  # process covariance matrix

        # These 2 attributes are the only state that changes during
        # filtering:
        self.xhat_k1 = initial_x  # a posteriori state estimate from step (k-1)
        self.P_k1 = initial_P    # a posteriori error estimate from step (k-1)

        self.ss = ss
        self.os = os
        self.AT = self.A.T
        self.CT = self.C.T

        if len(initial_x) != self.ss:
            raise ValueError('initial_x must be a vector with ss components')

    def step(self, y=None, isinitial=False, full_output=False, **kw):
        xhatminus, Pminus = self.step1__calculate_a_priori(isinitial=isinitial)
        return self.step2__calculate_a_posteri(xhatminus, Pminus, y=y,
                                               full_output=full_output, **kw)

    def step1__calculate_a_priori(self, isinitial=False):
        dot = numpy.dot  # shorthand
        ############################################
        #          update state-space

        # compute a priori estimate of statespace
        if not isinitial:
            xhatminus = dot(self.A, self.xhat_k1)
            # compute a priori estimate of errors
            Pminus = dot(dot(self.A, self.P_k1), self.AT) + self.Q
        else:
            xhatminus = self.xhat_k1
            Pminus = self.P_k1

        return xhatminus, Pminus

    def step2__calculate_a_posteri(self, xhatminus, Pminus, y=None,
                                   full_output=False, R=None):
        """
        y represents the observation for this time-step
        """
        if R is None:
            raise ValueError('R cannot be None')
        dot = numpy.dot  # shorthand
        inv = numpy.linalg.inv

        missing_data = False
        if y is None or numpy.any(numpy.isnan(y)):
            missing_data = True

        if not missing_data:
            ############################################
            #          incorporate observation

            # calculate a posteriori state estimate

            # calculate Kalman gain
            Knumerator = dot(Pminus, self.CT)
            Kdenominator = dot(dot(self.C, Pminus), self.CT) + R
            K = dot(Knumerator, inv(Kdenominator))  # Kalman gain

            residuals = y - dot(self.C, xhatminus)  # error/innovation
            xhat = xhatminus + dot(K, residuals)

            one_minus_KC = numpy.eye(self.ss) - dot(K, self.C)

            # compute a posteriori estimate of errors
            P = dot(one_minus_KC, Pminus)
        else:
            # no observation
            xhat = xhatminus
            P = Pminus

        if full_output:
            if missing_data:
                # XXX missing data, check literature!
                raise NotImplementedError(
                    "don't know how to compute VVnew with missing data")
                # VVnew = dot(self.A,self.P_k1)
                # loglik = 0
            else:
                # calculate loglik and Pfuture
                VVnew = dot(one_minus_KC, dot(self.A, self.P_k1))
                loglik = gaussian_prob(residuals,
                                       numpy.zeros((1, len(residuals))),
                                       Kdenominator, use_log=True)

        # this step (k) becomes next step's prior (k-1)
        self.xhat_k1 = xhat
        self.P_k1 = P
        if full_output:
            return xhat, P, loglik, VVnew
        else:
            return xhat, P


class KalmanFilter(VariableObservationNoiseKalmanFilter):

    def __init__(self, A, C, Q, R, initial_x, initial_P):
        self.R = R  # measurement covariance matrix
        VariableObservationNoiseKalmanFilter.__init__(
            self, A=A, C=C, Q=Q, initial_x=initial_x, initial_P=initial_P)
        assert R.shape == (self.os, self.os)

    def step2__calculate_a_posteri(self, xhatminus, Pminus, y=None,
                                   full_output=False):
        return VariableObservationNoiseKalmanFilter.step2__calculate_a_posteri(
            self, xhatminus=xhatminus, Pminus=Pminus, y=y, full_output=full_output,
            R=self.R)


def kalman_filter(y, A, C, Q, R, init_x, init_V, full_output=False):
    T = len(y)
    ss = len(A)
    R = numpy.array(R)

    for arr in (A, C, Q):
        if numpy.any(numpy.isnan(arr)):
            raise ValueError(
                "cannot do Kalman filtering with nan values in parameters")

    if R.ndim not in (2, 3):
        raise ValueError("R not 2 or 3 dimensions but %d" % R.ndim)

    if R.ndim == 2:
        kfilt = KalmanFilter(A, C, Q, R, init_x, init_V)
    else:
        assert R.ndim == 3
        if R.shape[0] != T:
            raise ValueError(
                'Per-observation noise must have same length as observations')
        kfilt = VariableObservationNoiseKalmanFilter(A, C, Q, init_x, init_V)

    # Forward pass
    xfilt = numpy.zeros((T, ss))
    Vfilt = numpy.zeros((T, ss, ss))
    if full_output:
        VVfilt = numpy.zeros((T, ss, ss))
    loglik = 0

    for i in range(T):
        isinitial = i == 0
        y_i = y[i]
        if R.ndim == 3:
            kw = dict(R=R[i])
        else:
            kw = {}

        if full_output:
            xfilt_i, Vfilt_i, LL, VVfilt_i = kfilt.step(y=y_i,
                                                        isinitial=isinitial,
                                                        full_output=True, **kw)
            VVfilt[i] = VVfilt_i
            loglik += LL
        else:
            xfilt_i, Vfilt_i = kfilt.step(y=y_i,
                                          isinitial=isinitial,
                                          full_output=False, **kw)
        xfilt[i] = xfilt_i
        Vfilt[i] = Vfilt_i

    if full_output:
        return xfilt, Vfilt, VVfilt, loglik
    else:
        return xfilt, Vfilt


def kalman_smoother(y, A, C, Q, R, init_x, init_V, valid_data_idx=None,
                    full_output=False):
    """Rauch-Tung-Striebel (RTS) smoother

    arguments
    ---------
    y - observations
    A - process update matrix
    C - state-to-observation matrix
    Q - process covariance matrix
    R - observation covariance matrix
    init_x - initial state
    init_V - initial error estimate
    valid_data_idx - (optional) Indices to rows of y that are valid or
        boolean array of len(y). (None if all data valid.)  Note that
        this is not necessary if y is nan where data are invalid.

    returns
    -------
    xsmooth - smoothed state estimates
    Vsmooth - smoothed error estimates
    VVsmooth - (only when full_output==True)
    loglik - (only when full_output==True)

    Kalman smoother based on Kevin Murphy's Kalman toolbox for
    MATLAB(tm).

    N.B. Axes are swapped relative to Kevin Murphy's example, because
    in all my data, time is the first dimension."""

    if valid_data_idx is not None:
        if hasattr(valid_data_idx, 'dtype') and valid_data_idx.dtype == numpy.bool:
            assert len(valid_data_idx) == len(y)
            invalid_cond = ~valid_data_idx
            y[invalid_cond] = numpy.nan  # broadcast
        else:
            y = numpy.array(y, copy=True)
            valid_data_idx = set(valid_data_idx)
            all_idx = set(range(len(y)))
            bad_idx = list(all_idx - valid_data_idx)
            for i in bad_idx:
                y[i] = numpy.nan  # broadcast

    def smooth_update(xsmooth_future, Vsmooth_future, xfilt, Vfilt,
                      Vfilt_future, VVfilt_future, A, Q, full_output=False):
        dot = numpy.dot
        inv = numpy.linalg.inv

        xpred = dot(A, xfilt)
        Vpred = dot(A, numpy.dot(Vfilt, A.T)) + Q
        J = dot(Vfilt, numpy.dot(A.T, inv(Vpred)))  # smoother gain matrix
        xsmooth = xfilt + dot(J, xsmooth_future - xpred)
        Vsmooth = Vfilt + dot(J, dot(Vsmooth_future - Vpred, J.T))
        VVsmooth_future = VVfilt_future + numpy.dot(
            (Vsmooth_future - Vfilt_future),
            numpy.dot(inv(Vfilt_future), VVfilt_future))
        return xsmooth, Vsmooth, VVsmooth_future

    T = len(y)
    ss = len(A)

    # Forward pass
    forward_results = kalman_filter(y, A, C, Q, R, init_x, init_V,
                                    full_output=full_output)
    if full_output:
        xfilt, Vfilt, VVfilt, loglik = forward_results
    else:
        xfilt, Vfilt = forward_results
        VVfilt = Vfilt  # dummy value

    # Backward pass
    xsmooth = numpy.array(xfilt, copy=True)
    Vsmooth = numpy.array(Vfilt, copy=True)
    VVsmooth = numpy.empty(Vfilt.shape)

    for t in range(T - 2, -1, -1):
        xsmooth_t, Vsmooth_t, VVsmooth_t = smooth_update(xsmooth[t + 1, :],
                                                         Vsmooth[t + 1, :, :],
                                                         xfilt[t, :],
                                                         Vfilt[t, :, :],
                                                         Vfilt[t + 1, :, :],
                                                         VVfilt[t + 1, :, :],
                                                         A, Q,
                                                         full_output=full_output)
        xsmooth[t, :] = xsmooth_t
        Vsmooth[t, :, :] = Vsmooth_t
        VVsmooth[t + 1, :, :] = VVsmooth_t

    VVsmooth[0, :, :] = numpy.zeros((ss, ss))

    if full_output:
        return xsmooth, Vsmooth, VVsmooth, loglik
    else:
        return xsmooth, Vsmooth



from setuptools import setup
import os, sys

import adskalman.version
version = adskalman.version.__version__

setup(name='adskalman',
      version=version,
      description='Kalman filtering routine',
      author_email='strawman@astraw.com',
      license='BSD',
      zip_safe=True,
      packages = ['adskalman'],
      package_data={'adskalman':['table*.csv',
                                 '*.mat'
                                 ]},
      )
